﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MbUnit.Framework;
using DataTransfer;

namespace DataAccessLayerTest
{
    [TestFixture]
    public class NHibernateDataProviderTest : Microdesk.Utility.UnitTest.DatabaseUnitTestBase
    {
        #region Fields

        private DataAccessLayer.NHibernateDataProvider _provider;

        private NHibernate.ISession _session;

        private DataAccessLayer.NHibernateSessionManager _sessionManager;



        private void CreateDatabase()
        {
            NHibernate.Cfg.Configuration cfg = new NHibernate.Cfg.Configuration();

            cfg.Configure();

            NHibernate.Tool.hbm2ddl.SchemaExport schema = new NHibernate.Tool.hbm2ddl.SchemaExport(cfg);

            schema.Execute(true, true, false, true);

        }



        [Test]
        public void CanGetCustomersOrderingYoYosSince()
        {
            DateTime testorderDate = new DateTime(2008, 2, 1);
            IList<YoYoCustomer> customers = _provider.GetCustomersOrderingYoYosSince(testorderDate);

            foreach (YoYoCustomer yoyo in customers)
            {
                Assert.GreaterThan(yoyo.OrderDate, testorderDate);
            }


        }

        [Test]
        public void NameComponenetCanReturnFullname()
        {
            const string CUSTOMER_FIRSTNAME = "Steve";
            const string CUSTOMER_LASTNAME = "Bohlen";
            const string CUSTOMER_FULLNAME = "Steve Bohlen";

            IList<Customer> customers = _provider.GetCustomersByFirstnameAndLastname(CUSTOMER_FIRSTNAME, CUSTOMER_LASTNAME);


            foreach (Customer c in customers)
            {

                Assert.AreEqual(CUSTOMER_FULLNAME, c.Name.Fullname);
            }


        }


        #endregion

        #region Public Methods

        [Test]
        [ExpectedException(typeof(NHibernate.HibernateException))]
        public void AddCustomerThrowsExceptionOnFail()
        {
            _provider.AddCustomer(BuildInvalidCustomer());
        }

        [Test]
        public void CanAddCustomer()
        {
            Customer customer = new Customer() { Name = new Name() { Firstname = "Steve", Lastname = "Bohlen" } };

            int newIdentity = _provider.AddCustomer(customer);

            Customer testCustomer = _provider.GetCustomerById(newIdentity);

            Assert.IsNotNull(testCustomer);

        }

        [Test]
        public void CanDeleteCustomer()
        {
            Customer customer = GetDeletableCustomer();

            _provider.DeleteCustomer(customer);

            Customer testCustomer = GetDeletableCustomer();

            Assert.IsNull(testCustomer);
        }

        [Test]
        [Ignore]
        public void CanDemonstrateLazyLoadingSQLSequence()
        {
            System.Threading.Thread.Sleep(10000);

            Customer customer = _provider.GetCustomerByIdAndLeaveSessionIntact(1);

            System.Threading.Thread.Sleep(10000);

            int numOrders = customer.Orders.Count;
        }

        [Test]
        public void CanGetCountsOfCustomerFirsnames()
        {
            IList<CustomerFirstnameCounter> firstnameCounts = _provider.GetCustomerFirstnameCounts();

            Dictionary<string, int> expectedCounts = new Dictionary<string, int>();
            expectedCounts.Add("Steve", 3);
            expectedCounts.Add("Mike", 1);
            expectedCounts.Add("Peter", 1);
            expectedCounts.Add("Joe", 1);
            expectedCounts.Add("Yuliana", 1);

            foreach (CustomerFirstnameCounter item in firstnameCounts)
            {
                Assert.AreEqual(expectedCounts[item.Firstname], item.Count);
                Assert.IsTrue(expectedCounts.ContainsKey(item.Firstname));
            }

        }

        [Test]
        public void CanGetCustomerAndOrdersByCustomerId()
        {
            Customer customer = _provider.GetCustomerById(1);
            Assert.AreEqual(1, customer.CustomerId);

            Assert.GreaterThan(customer.Orders.Count, 0);
        }

        [Test]
        public void CanGetCustomerByExample()
        {
            const string CUSTOMER_FIRSTNAME = "Steve";
            const int QTY_CUSTOMERS_EXPECTED = 2;
            const string CUSTOMER_LASTNAME = "Bohlen";
            Customer theCustomer = new Customer();

            Name name = new Name() { Firstname = CUSTOMER_FIRSTNAME, Lastname = CUSTOMER_LASTNAME };

            theCustomer.Name = name;

            IList<Customer> customers = _provider.GetCustomersByExample(theCustomer);

            if (customers.Count == 0)
                Assert.Warning("YOU MAY NOT BE TESTING WHAT YOU THINK YOU ARE!");

            foreach (Customer c in customers)
            {
                Assert.AreEqual(CUSTOMER_FIRSTNAME, c.Name.Firstname);
                Assert.AreEqual(CUSTOMER_LASTNAME, c.Name.Lastname);
            }

            Assert.AreEqual(QTY_CUSTOMERS_EXPECTED, customers.Count);
        }

        [Test]
        public void CanGetCustomerByFirstname()
        {
            const string CUSTOMER_FIRSTNAME = "Steve";
            const int QTY_CUSTOMERS_EXPECTED = 3;

            IList<Customer> customers = _provider.GetCustomersByFirstname(CUSTOMER_FIRSTNAME);

            if (customers.Count == 0)
                Assert.Warning("YOU MAY NOT BE TESTING WHAT YOU THINK YOU ARE!");

            foreach (Customer c in customers)
            {
                Assert.AreEqual(CUSTOMER_FIRSTNAME, c.Name.Firstname);
            }

            Assert.AreEqual(QTY_CUSTOMERS_EXPECTED, customers.Count);

        }

        [Test]
        public void CanGetCustomerByFirstnameAndLastname()
        {
            const string CUSTOMER_FIRSTNAME = "Steve";
            const string CUSTOMER_LASTNAME = "Bohlen";
            const int QTY_CUSTOMER_EXPECTED = 2;

            IList<Customer> customers = _provider.GetCustomersByFirstnameAndLastname(CUSTOMER_FIRSTNAME, CUSTOMER_LASTNAME);

            if (customers.Count == 0)
                Assert.Warning("YOU MAY NOT BE TESTING WHAT YOU THINK YOU ARE!");

            foreach (Customer c in customers)
            {
                Assert.AreEqual(CUSTOMER_FIRSTNAME, c.Name.Firstname);
                Assert.AreEqual(CUSTOMER_LASTNAME, c.Name.Lastname);
            }

            Assert.AreEqual(QTY_CUSTOMER_EXPECTED, customers.Count);
        }

        [Test]
        public void CanGetCustomerById()
        {
            Customer customer = _provider.GetCustomerById(1);
            Assert.AreEqual(1, customer.CustomerId);
        }

        [Test]
        public void CanGetCustomersOrderedByLastname()
        {
            IList<Customer> customers = _provider.GetCustomersOrderedByLastnames();

            Customer priorCustomer = null;

            foreach (Customer c in customers)
            {
                if (priorCustomer != null)
                    Assert.GreaterEqualThan(string.Compare(c.Name.Lastname, priorCustomer.Name.Lastname), 0);

                priorCustomer = c;
            }
        }

        [Test]
        public void CanGetCustomersWithIdGreaterThan()
        {
            IList<Customer> customers = _provider.GetCustomersWithIdGreaterThan(20);

            if (customers.Count == 0)
                Assert.Warning("YOU MAY NOT BE TESTING WHAT YOU THINK YOU ARE!");

            foreach (Customer c in customers)
            {
                Assert.GreaterThan(c.CustomerId, 20);
            }
        }

        [Test]
        public void CanGetCustomersWithOrdersHavingProduct()
        {
            IList<Customer> customers = _provider.GetCustomersWithOrdersHavingProduct(1);

            bool found = false; ;

            foreach (Customer c in customers)
            {
                foreach (Order o in c.Orders)
                {
                    foreach (Product p in o.Products)
                    {
                        if (p.ProductId == 1)
                        {
                            found = true;
                            break;
                        }
                    }
                }
            }

            Assert.IsTrue(found);

        }

        [Test]
        public void CanGetDistinctCustomerFirstnames()
        {
            IList<string> firstnames = _provider.GetDistinctCustomerFirstnames();

            if (firstnames.Count == 0)
                Assert.Warning("No Firstnames were found!");

            foreach (string fn in firstnames)
            {
                Assert.AreEqual(1, firstnames.Count<string>(x => x == fn));
            }

        }

        [Test]
        public void CanGetDistinctCustomersWithOrdersSince()
        {
            DateTime testOrderDate = new DateTime(2008, 4, 1);

            IList<Customer> customers = _provider.GetDistinctCustomersWithOrdersSince(testOrderDate);

            foreach (Customer c in customers)
            {
                foreach (Order o in c.Orders)
                {
                    Assert.GreaterThan(o.OrderDate, testOrderDate);
                }
            }

            foreach (Customer c in customers)
            {
                Assert.AreEqual(1, customers.Count<Customer>(x => x == c));
            }

        }

        [Test]
        public void CanSaveOrUpdateCustomers()
        {
            const string CUSTOMER_FIRSTNAME = "Steve";
            IList<Customer> customers = _provider.GetCustomersByFirstname(CUSTOMER_FIRSTNAME);

            if (customers.Count == 0)
                Assert.Warning("You aren't testing what you think you are here!");

            const string NEW_CUSTOMER_LASTNAME = "Goodbye";
            foreach (Customer c in customers)
            {
                c.Name.Lastname = NEW_CUSTOMER_LASTNAME;
            }

            Customer c1 = new Customer() { Name = new Name() { Firstname = CUSTOMER_FIRSTNAME, Lastname = NEW_CUSTOMER_LASTNAME } };
            Customer c2 = new Customer() { Name = new Name() { Firstname = CUSTOMER_FIRSTNAME, Lastname = NEW_CUSTOMER_LASTNAME } };

            customers.Add(c1);
            customers.Add(c2);

            int initialListCount = customers.Count;

            _provider.SaveOrUpdateCustomers(customers);

            int testListCount = _provider.GetCustomersByFirstnameAndLastname(CUSTOMER_FIRSTNAME, NEW_CUSTOMER_LASTNAME).Count;

            Assert.AreEqual(initialListCount, testListCount);

        }

        [Test]
        [ExpectedException(typeof(NHibernate.StaleObjectStateException))]
        public void CanThrowExceptionOnConcurrencyViolation()
        {

            Customer c1;
            using (_session)
            {
                c1 = _provider.GetCustomerById(1);
            }

            ResetSessionForProvider();

            Customer c2;

            using (_session)
            {
                c2 = _provider.GetCustomerById(1);
            }
            ResetSessionForProvider();

            Name c1name = new Name() { Firstname = "hello", Lastname = "goodbye" };


            c1.Name = c1name;

            using (_session)
            {
                _provider.UpdateCustomer(c1);
            }
            ResetSessionForProvider();

            using (_session)
            {
                _provider.UpdateCustomer(c2);
            }

        }

        [Test]
        [ExpectedException(typeof(NHibernate.StaleObjectStateException))]
        public void CanThrowExceptionOnConcurrencyViolationOnDelete()
        {
            Customer c1;
            using (_session)
            {
                c1 = GetDeletableCustomer();
            }

            ResetSessionForProvider();

            Customer c2;
            using (_session)
            {
                c2 = GetDeletableCustomer();
            }

            ResetSessionForProvider();

            using (_session)
            {
                _provider.DeleteCustomer(c1);
            }

            ResetSessionForProvider();

            using (_session)
            {
                _provider.DeleteCustomer(c2);
            }
        }

        [Test]
        public void CanUpdateCustomer()
        {
            Customer customer = _provider.GetCustomerById(1);

            string originalFirstname = customer.Name.Firstname;

            string newFirstname = string.Concat(originalFirstname, "SUFFIX");

            customer.Name.Firstname = newFirstname;

            _provider.UpdateCustomer(customer);

            Customer testCustomer = _provider.GetCustomerById(1);

            Assert.AreEqual(newFirstname, testCustomer.Name.Firstname);

        }

        [Test]
        public void CRIT_CanGetCustomerByFirstname()
        {
            const string CUSTOMER_FIRSTNAME = "Steve";
            const int QTY_CUSTOMERS_EXPECTED = 3;

            IList<Customer> customers = _provider.CRIT_GetCustomersByFirstname(CUSTOMER_FIRSTNAME);

            if (customers.Count == 0)
                Assert.Warning("YOU MAY NOT BE TESTING WHAT YOU THINK YOU ARE!");

            foreach (Customer c in customers)
            {
                Assert.AreEqual(CUSTOMER_FIRSTNAME, c.Name.Firstname);
            }

            Assert.AreEqual(QTY_CUSTOMERS_EXPECTED, customers.Count);

        }

        [Test]
        public void CRIT_CanGetPreferredCustomerByFirstname()
        {
            const string CUSTOMER_FIRSTNAME = "Steve";
            const int QTY_CUSTOMERS_EXPECTED = 1;

            IList<PreferredCustomer> customers = _provider.CRIT_GetPreferredCustomersByFirstname(CUSTOMER_FIRSTNAME);

            if (customers.Count == 0)
                Assert.Warning("YOU MAY NOT BE TESTING WHAT YOU THINK YOU ARE!");

            foreach (PreferredCustomer c in customers)
            {
                Assert.AreEqual(CUSTOMER_FIRSTNAME, c.Name.Firstname);
            }

            Assert.AreEqual(QTY_CUSTOMERS_EXPECTED, customers.Count);

        }

        [Test]
        public void CRIT_CanGetCustomerByFirstnameAndLastname()
        {
            const string CUSTOMER_FIRSTNAME = "Steve";
            const string CUSTOMER_LASTNAME = "Bohlen";
            const int QTY_CUSTOMER_EXPECTED = 2;

            IList<Customer> customers = _provider.CRIT_GetCustomersByFirstnameAndLastname(CUSTOMER_FIRSTNAME, CUSTOMER_LASTNAME);

            if (customers.Count == 0)
                Assert.Warning("YOU MAY NOT BE TESTING WHAT YOU THINK YOU ARE!");

            foreach (Customer c in customers)
            {
                Assert.AreEqual(CUSTOMER_FIRSTNAME, c.Name.Firstname);
                Assert.AreEqual(CUSTOMER_LASTNAME, c.Name.Lastname);
            }

            Assert.AreEqual(QTY_CUSTOMER_EXPECTED, customers.Count);
        }

        [Test]
        public void CRIT_CanGetCustomersOrderedByLastname()
        {
            IList<Customer> customers = _provider.CRIT_GetCustomersOrderedByLastnames();

            Customer priorCustomer = null;

            foreach (Customer c in customers)
            {
                if (priorCustomer != null)
                    Assert.GreaterEqualThan(string.Compare(c.Name.Lastname, priorCustomer.Name.Lastname), 0);

                priorCustomer = c;
            }
        }

        [Test]
        public void CRIT_CanGetCustomersWithIdGreaterThan()
        {
            IList<Customer> customers = _provider.CRIT_GetCutomersWithIdGreaterThan(20);

            if (customers.Count == 0)
                Assert.Warning("YOU MAY NOT BE TESTING WHAT YOU THINK YOU ARE!");

            foreach (Customer c in customers)
            {
                Assert.GreaterThan(c.CustomerId, 20);
            }
        }

        [Test]
        public void CRIT_CanGetDistinctCustomerFirstnames()
        {
            IList<string> firstnames = _provider.CRIT_GetDistinctCustomerFirstnames();

            if (firstnames.Count == 0)
                Assert.Warning("No Firstnames were found!");

            foreach (string fn in firstnames)
            {
                Assert.AreEqual(1, firstnames.Count<string>(x => x == fn));
            }

        }

        [Test]
        public void CRIT_CanGetDistinctCustomersWithOrdersSince()
        {
            DateTime testOrderDate = new DateTime(2008, 4, 1);

            IList<Customer> customers = _provider.CRIT_GetDistinctCustomersWithOrdersSince(testOrderDate);

            foreach (Customer c in customers)
            {
                foreach (Order o in c.Orders)
                {
                    Assert.GreaterThan(o.OrderDate, testOrderDate);
                }
            }

            foreach (Customer c in customers)
            {
                Assert.AreEqual(1, customers.Count<Customer>(x => x == c));
            }

        }

        [Test]
        [ExpectedException(typeof(NHibernate.HibernateException))]
        public void DeleteCustomerCanThrowExceptionOnFail()
        {
            Customer customer = GetUndeletableCustomer();

            _provider.DeleteCustomer(customer);
        }

        public void GetMyTestDataXMLFile()
        {
            SaveTestDatabase();
        }

        [Test]
        public void SaveOrUpdateCustomersThrowsExceptionOnFail()
        {
            const string CUSTOMER_FIRSTNAME = "Steve";
            IList<Customer> customers = _provider.GetCustomersByFirstname(CUSTOMER_FIRSTNAME);

            if (customers.Count == 0)
                Assert.Warning("You aren't testing what you think you are here!");

            const string NEW_CUSTOMER_LASTNAME = "Goodbye";
            foreach (Customer c in customers)
            {
                c.Name.Lastname = NEW_CUSTOMER_LASTNAME;
            }

            Customer c1 = BuildInvalidCustomer();
            Customer c2 = new Customer() { Name = new Name() { Firstname = CUSTOMER_FIRSTNAME, Lastname = NEW_CUSTOMER_LASTNAME } };

            customers.Add(c1);
            customers.Add(c2);

            int initialListCount = customers.Count;

            try
            {
                _provider.SaveOrUpdateCustomers(customers);
            }
            catch (NHibernate.HibernateException)
            {
                ResetSessionForProvider();
            }

            int testListCount = _provider.GetCustomersByFirstnameAndLastname(CUSTOMER_FIRSTNAME, NEW_CUSTOMER_LASTNAME).Count;

            Assert.AreEqual(0, testListCount);

        }

        [SetUp]
        public void Setup()
        {
            DatabaseSetUp();

            _session = _sessionManager.GetSession();
            _provider = new DataAccessLayer.NHibernateDataProvider(_session);
        }

        [TearDown]
        public void TearDown()
        {
            DatabaseTearDown();
        }

        [TestFixtureSetUp]
        public void TestFixtureSetup()
        {
            DatabaseFixtureSetUp();

            _sessionManager = new DataAccessLayer.NHibernateSessionManager();
        }

        [TestFixtureTearDown]
        public void TestFixtureTearDown()
        {
            DatabaseFixtureTearDown();
        }

        [Test]
        [ExpectedException(typeof(NHibernate.HibernateException))]
        public void UpdateCustomerCanThrowExceptionOnFail()
        {
            Customer customer = _provider.GetCustomerById(1);

            customer.Name.Firstname = "012345678901234567890123456789012345678901234567890123456789";

            _provider.UpdateCustomer(customer);
        }

        #endregion

        #region Private Methods

        private Customer BuildInvalidCustomer()
        {
            return new Customer() { Name = new Name() { Firstname = "012345678901234567890123456789012345678901234567890123456789" } };
        }

        private Customer GetDeletableCustomer()
        {
            return _provider.GetCustomerById(3);
        }

        private Customer GetUndeletableCustomer()
        {
            return _provider.GetCustomerById(1);
        }

        private void ResetSessionForProvider()
        {
            _session = _sessionManager.GetSession();
            _provider.Session = _session;
        }

        #endregion

    }
}
